@@ -8,6 +8,7 @@ struct message_size;
 #include "guid.h"
 #include "mail-types.h"
 #include "mail-error.h"
+#include "mail-index.h"
 #include "mail-namespace.h"
 #include "mailbox-list.h"
 #include "mailbox-attribute.h"
@@ -15,6 +16,9 @@ struct message_size;
 /* If some operation is taking long, call notify_ok every n seconds. */
 #define MAIL_STORAGE_STAYALIVE_SECS 15
 
+#define MAIL_KEYWORD_HAS_ATTACHMENT "$HasAttachment"
+#define MAIL_KEYWORD_HAS_NO_ATTACHMENT "$HasNoAttachment"
+
 enum mail_storage_flags {
 	/* Remember message headers' MD5 sum */
 	MAIL_STORAGE_FLAG_KEEP_HEADER_MD5	= 0x01,
@@ -61,13 +65,22 @@ enum mailbox_flags {
 	   The backend shouldn't treat it as corruption if a mail body isn't
 	   found. */
 	MAILBOX_FLAG_USE_STUBS		= 0x800,
+	/* Mailbox is created implicitly if it does not exist. */
+	MAILBOX_FLAG_AUTO_CREATE	= 0x1000,
+	/* Mailbox is subscribed to implicitly when it is created automatically */
+	MAILBOX_FLAG_AUTO_SUBSCRIBE	= 0x2000,
+	/* Run fsck for mailbox index before doing anything else. This may be
+	   useful in fixing index corruption errors that aren't otherwise
+	   detected and that are causing the full mailbox opening to fail. */
+	MAILBOX_FLAG_FSCK		= 0x4000,
+	/* Interpret name argument for mailbox_alloc_for_user() as a SPECIAL-USE
+	   flag. */
+	MAILBOX_FLAG_SPECIAL_USE	= 0x8000,
 };
 
 enum mailbox_feature {
 	/* Enable tracking modsequences */
 	MAILBOX_FEATURE_CONDSTORE	= 0x01,
-	/* Enable tracking expunge modsequences */
-	MAILBOX_FEATURE_QRESYNC		= 0x02
 };
 
 enum mailbox_existence {
@@ -164,7 +177,6 @@ enum mail_fetch_field {
 	MAIL_FETCH_FROM_ENVELOPE	= 0x00008000,
 	MAIL_FETCH_HEADER_MD5		= 0x00010000,
 	MAIL_FETCH_STORAGE_ID		= 0x00020000,
-#define MAIL_FETCH_UIDL_FILE_NAME MAIL_FETCH_STORAGE_ID /* FIXME: remove in v2.3 */
 	MAIL_FETCH_UIDL_BACKEND		= 0x00040000,
 	MAIL_FETCH_MAILBOX_NAME		= 0x00080000,
 	MAIL_FETCH_SEARCH_RELEVANCY	= 0x00100000,
@@ -263,21 +275,21 @@ struct mailbox_status {
 	enum mail_flags flags;
 
 	/* All keywords can be permanently modified (STATUS_PERMANENT_FLAGS) */
-	unsigned int permanent_keywords:1;
+	bool permanent_keywords:1;
 	/* More keywords can be created (STATUS_PERMANENT_FLAGS) */
-	unsigned int allow_new_keywords:1;
+	bool allow_new_keywords:1;
 	/* Modseqs aren't permanent (index is in memory) (STATUS_HIGHESTMODSEQ) */
-	unsigned int nonpermanent_modseqs:1;
+	bool nonpermanent_modseqs:1;
 	/* Modseq tracking has never been enabled for this mailbox
 	   yet. (STATUS_HIGHESTMODSEQ) */
-	unsigned int no_modseq_tracking:1;
+	bool no_modseq_tracking:1;
 
 	/* Messages have GUIDs (always set) */
-	unsigned int have_guids:1;
+	bool have_guids:1;
 	/* mailbox_save_set_guid() works (always set) */
-	unsigned int have_save_guids:1;
+	bool have_save_guids:1;
 	/* GUIDs are always 128bit (always set) */
-	unsigned int have_only_guid128:1;
+	bool have_only_guid128:1;
 };
 
 struct mailbox_cache_field {
@@ -336,8 +348,8 @@ struct mail_transaction_commit_changes {
 	/* number of modseq changes that couldn't be changed as requested */
 	unsigned int ignored_modseq_changes;
 
-	/* TRUE if anything actually changed with this commit */
-	bool changed;
+	/* Changes that occurred within this transaction */
+	enum mail_index_transaction_change changes_mask;
 	/* User doesn't have read ACL for the mailbox, so don't show the
 	   uid_validity / saved_uids. */
 	bool no_read_perm;
@@ -349,7 +361,7 @@ struct mailbox_sync_rec {
 };
 struct mailbox_sync_status {
 	/* There are expunges that haven't been synced yet */
-	unsigned int sync_delayed_expunges:1;
+	bool sync_delayed_expunges:1;
 };
 
 struct mailbox_expunge_rec {
@@ -370,7 +382,12 @@ enum mail_lookup_abort {
 	   operations to be performed, such as stat()ing a file. */
 	MAIL_LOOKUP_ABORT_READ_MAIL,
 	/* Abort if the operation can't be done fully using cache file */
-	MAIL_LOOKUP_ABORT_NOT_IN_CACHE
+	MAIL_LOOKUP_ABORT_NOT_IN_CACHE,
+	/* Abort if the operation can't be done fully using cache file.
+	 * During this lookup all cache lookups that have "no" decision
+	 * will be changed to "tmp". This way the field will start to be
+	 * cached in the future. */
+	MAIL_LOOKUP_ABORT_NOT_IN_CACHE_START_CACHING,
 };
 
 enum mail_access_type {
@@ -385,12 +402,13 @@ struct mail {
 	/* always set */
 	struct mailbox *box;
 	struct mailbox_transaction_context *transaction;
+	struct event *event;
 	uint32_t seq, uid;
 
-	unsigned int expunged:1;
-	unsigned int saving:1; /* This mail is still being saved */
-	unsigned int has_nuls:1; /* message data is known to contain NULs */
-	unsigned int has_no_nuls:1; /* -''- known to not contain NULs */
+	bool expunged:1;
+	bool saving:1; /* This mail is still being saved */
+	bool has_nuls:1; /* message data is known to contain NULs */
+	bool has_no_nuls:1; /* -''- known to not contain NULs */
 
 	/* Mail's header/body stream was opened within this request.
 	   If lookup_abort!=MAIL_LOOKUP_ABORT_NEVER, this can't become TRUE. */
@@ -481,6 +499,14 @@ mailbox_get_last_error(struct mailbox *box, enum mail_error *error_r)
 /* Wrapper for mail_storage_get_last_error(); */
 enum mail_error mailbox_get_last_mail_error(struct mailbox *box);
 
+const char * ATTR_NOWARN_UNUSED_RESULT
+mail_storage_get_last_internal_error(struct mail_storage *storage,
+				     enum mail_error *error_r) ATTR_NULL(2);
+/* Wrapper for mail_storage_get_last_internal_error(); */
+const char * ATTR_NOWARN_UNUSED_RESULT
+mailbox_get_last_internal_error(struct mailbox *box,
+				enum mail_error *error_r) ATTR_NULL(2);
+
 /* Save the last error until it's popped. This is useful for cases where the
    storage has already failed, but the cleanup code path changes the error to
    something else unwanted. */
@@ -499,6 +525,17 @@ struct mailbox *mailbox_alloc(struct mailbox_list *list, const char *vname,
 struct mailbox *mailbox_alloc_guid(struct mailbox_list *list,
 				   const guid_128_t guid,
 				   enum mailbox_flags flags);
+/* Initialize mailbox for a particular user without actually opening any files
+   or verifying that it exists. The mname parameter is normally equal to the
+   mailbox vname, except when the MAILBOX_FLAG_SPECIAL_USE flag is set, in which
+   case it is the special-use flag. */
+struct mailbox *
+mailbox_alloc_for_user(struct mail_user *user, const char *mname,
+		       enum mailbox_flags flags);
+
+/* Set a human-readable reason for why this mailbox is being accessed.
+   This is used for logging purposes. */
+void mailbox_set_reason(struct mailbox *box, const char *reason);
 /* Get mailbox existence state. If auto_boxes=FALSE, return
    MAILBOX_EXISTENCE_NONE for autocreated mailboxes that haven't been
    physically created yet */
@@ -521,6 +558,9 @@ bool mailbox_equals(const struct mailbox *box1,
 /* Returns TRUE if the mailbox is user's INBOX or another user's shared INBOX */
 bool mailbox_is_any_inbox(struct mailbox *box);
 
+/* Returns TRUE if the mailbox has the specified special use flag assigned. */
+bool mailbox_has_special_use(struct mailbox *box, const char *special_use);
+
 /* Change mailbox_verify_create_name() to not verify new mailbox name
    restrictions (but still check that it's a valid existing name). This is
    mainly used by dsync to make sure the sync works even though the original
@@ -625,13 +665,14 @@ void mailbox_notify_changes(struct mailbox *box,
 	ATTR_NULL(3);
 #define mailbox_notify_changes(box, callback, context) \
 	  mailbox_notify_changes(box, (mailbox_notify_callback_t *)callback, \
-		(void *)((char *)context + CALLBACK_TYPECHECK(callback, \
+		(void *)((char *)context - CALLBACK_TYPECHECK(callback, \
 			void (*)(struct mailbox *, typeof(context)))))
 void mailbox_notify_changes_stop(struct mailbox *box);
 
 struct mailbox_transaction_context *
 mailbox_transaction_begin(struct mailbox *box,
-			  enum mailbox_transaction_flags flags);
+			  enum mailbox_transaction_flags flags,
+			  const char *reason);
 int mailbox_transaction_commit(struct mailbox_transaction_context **t);
 int mailbox_transaction_commit_get_changes(
 	struct mailbox_transaction_context **t,
@@ -676,6 +717,10 @@ struct mailbox_header_lookup_ctx *
 mailbox_header_lookup_init(struct mailbox *box, const char *const headers[]);
 void mailbox_header_lookup_ref(struct mailbox_header_lookup_ctx *ctx);
 void mailbox_header_lookup_unref(struct mailbox_header_lookup_ctx **ctx);
+/* Merge two header lookups. */
+struct mailbox_header_lookup_ctx *
+mailbox_header_lookup_merge(const struct mailbox_header_lookup_ctx *hdr1,
+			    const struct mailbox_header_lookup_ctx *hdr2);
 
 /* Initialize new search request. If sort_program is non-NULL, the messages are
    returned in the requested order, otherwise from first to last. */
@@ -698,6 +743,10 @@ bool mailbox_search_next_nonblock(struct mail_search_context *ctx,
    determine correctly if those messages should have been returned in this
    search. */
 bool mailbox_search_seen_lost_data(struct mail_search_context *ctx);
+/* Detach the given mail from the search context. This allows the mail to live
+   even after mail_search_context has been freed. */
+void mailbox_search_mail_detach(struct mail_search_context *ctx,
+				struct mail *mail);
 
 /* Remember the search result for future use. This must be called before the
    first mailbox_search_next*() call. */
@@ -733,6 +782,10 @@ mailbox_keywords_create_valid(struct mailbox *box,
 struct mail_keywords *
 mailbox_keywords_create_from_indexes(struct mailbox *box,
 				     const ARRAY_TYPE(keyword_indexes) *idx);
+/* Return union of two mail_keywords. They must be created in the same
+   mailbox. */
+struct mail_keywords *mailbox_keywords_merge(struct mail_keywords *keywords1,
+					     struct mail_keywords *keywords2);
 void mailbox_keywords_ref(struct mail_keywords *keywords);
 void mailbox_keywords_unref(struct mail_keywords **keywords);
 /* Returns TRUE if keyword is valid, FALSE and error if not. */
@@ -779,9 +832,6 @@ void mailbox_save_set_pop3_uidl(struct mail_save_context *ctx,
    of the mailbox. Not all backends support this. */
 void mailbox_save_set_pop3_order(struct mail_save_context *ctx,
 				 unsigned int order);
-/* FIXME: Remove in v2.3. Obsolete - use mailbox_save_get_dest_mail() instead */
-void mailbox_save_set_dest_mail(struct mail_save_context *ctx,
-				struct mail *mail);
 /* Returns the destination mail */
 struct mail *mailbox_save_get_dest_mail(struct mail_save_context *ctx);
 /* Begin saving the message. All mail_save_set_*() calls must have been called
@@ -819,7 +869,10 @@ void mail_set_seq(struct mail *mail, uint32_t seq);
 bool mail_set_uid(struct mail *mail, uint32_t uid);
 
 /* Add wanted fields/headers on top of existing ones. These will be forgotten
-   after the next mail_set_seq/uid(). */
+   after the next mail_set_seq/uid() that closes the existing mail. Note that
+   it's valid to call this function while there is no mail assigned
+   (mail->seq==0), i.e. this is called before any mail_set_seq/uid() or after
+   mail.close(). */
 void mail_add_temp_wanted_fields(struct mail *mail,
 				 enum mail_fetch_field fields,
 				 struct mailbox_header_lookup_ctx *headers)
@@ -869,7 +922,8 @@ int mail_get_first_header(struct mail *mail, const char *field,
 int mail_get_first_header_utf8(struct mail *mail, const char *field,
 			       const char **value_r);
 /* Return a NULL-terminated list of values for each found field.
-   Returns 1 if headers were found, 0 if not (value_r==NULL) or -1 if error. */
+   Returns 1 if headers were found, 0 if not (value_r[0]==NULL) or
+   -1 if error. */
 int mail_get_headers(struct mail *mail, const char *field,
 		     const char *const **value_r);
 /* Like mail_get_headers(), but decode MIME encoded words to UTF-8.
@@ -923,9 +977,6 @@ int mail_get_special(struct mail *mail, enum mail_fetch_field field,
 /* Returns the mail for the physical message. Normally this is the mail itself,
    but in virtual mailboxes it points to the backend mailbox. */
 int mail_get_backend_mail(struct mail *mail, struct mail **real_mail_r);
-/* FIXME: For backwards compatibility for now, use mail_get_backend_mail()
-   instead. */
-struct mail *mail_get_real_mail(struct mail *mail);
 
 /* Update message flags. */
 void mail_update_flags(struct mail *mail, enum modify_type modify_type,
@@ -946,10 +997,9 @@ void mail_expunge(struct mail *mail);
 /* Add missing fields to cache. */
 void mail_precache(struct mail *mail);
 /* Mark a cached field corrupted and have it recalculated. */
-void mail_set_cache_corrupted(struct mail *mail, enum mail_fetch_field field);
-void mail_set_cache_corrupted_reason(struct mail *mail,
-				     enum mail_fetch_field field,
-				     const char *reason);
+void mail_set_cache_corrupted(struct mail *mail,
+			      enum mail_fetch_field field,
+			      const char *reason);
 
 /* Return 128 bit GUID using input string. If guid is already 128 bit hex
    encoded, it's returned as-is. Otherwise SHA1 sum is taken and its last
